#include <stdio.h>
#include <string.h>

#include "SoundBuffer.h"
#include "Constants.h"
#include "Timing.h"
#include "Log.h"

// holds our samples, one byte per sample
static Uint8* _buffer;
static Uint32 _bufferSize;

// tends to take longer to synchronize on faster computers/newer OSes
// EXPERIMENTALLY DETERMINED
static double _initialSyncPeriodSeconds = 3.0f;

static Uint32 _sampleRatePerSecond;

// these represent the read and write "heads"
//     SDL's audio module reads
//     CPU activity writes
static Uint32 _read_index;
static Uint32 _write_index;

// tracks total distance travelled by read and write "heads"
static Sint64 _read_index_distance_traveled;
static Sint64 _write_index_distance_traveled;

// counts all occurrences when resynchronization is needed, so it can be reported in the log
static Uint32 _resync_count__read_is_too_close_behind_write;
static Uint32 _resync_count__read_is_too_far_behind_write;

#define _DEBUG_BUFFER_SIZE 256
static const char* _debugBuffer[_DEBUG_BUFFER_SIZE];

// resynchronization parameters
Uint32 _read_minimum_trailing_distance;
Uint32 _read_maximum_trailing_distance;
Uint32 _read_sync_rewind_amount;
Uint32 _read_sync_fastforward_amount;

#define READ_MINIMUM_TRAILING_DISTANCE__FACTOR 20
#define READ_MAXIMUM_TRAILING_DISTANCE__FACTOR 150
// amounts above 10 cause constant stutter in manic miner
#define READ_SYNC_REWIND_AMOUNT__FACTOR        5
#define READ_SYNC_FASTFORWARD_AMOUNT__FACTOR   5

// rewinds (moves "backward") a circular buffer index by AT MOST one full buffer length
void _soundbuffer_rewind_index(Uint32* index, Uint64* indexDistanceTravelled, Uint32 rewindAmount) {
    if (*index >= rewindAmount) {
        *index -= rewindAmount;
    }
    else {
        *index = *index + _bufferSize - rewindAmount;
    }

    *indexDistanceTravelled -= rewindAmount;
}

// fast-forwards (moves "forward") a circular buffer index by AT MOST one full buffer length
void _soundbuffer_fastforward_index(Uint32* index, Uint64* indexDistanceTravelled, Uint32 fastForwardAmount) {
    *index += fastForwardAmount;
    if (*index >= _bufferSize) {
        *index -= _bufferSize;
    }

    *indexDistanceTravelled += fastForwardAmount;
}

void _soundbuffer_rewind_read() {
    _soundbuffer_rewind_index(&_read_index, &_read_index_distance_traveled, _read_sync_rewind_amount);
}

void _soundbuffer_fastforward_read() {
    _soundbuffer_fastforward_index(&_read_index, &_read_index_distance_traveled, _read_sync_fastforward_amount);
}

// depending on computer and OS, sound is poorly synchronized for a short while at application start
// this can report much higher resync numbers
Uint8 _soundbuffer_is_in_initial_sync_period() {
    Uint32 syncPeriodLengthInSamples = (Uint32)((double)_sampleRatePerSecond * _initialSyncPeriodSeconds);
    if (_read_index_distance_traveled > syncPeriodLengthInSamples) {
        // we've read past the initial sync period
        return 0;
    }

    return 1;
}

Uint8 soundbuffer_read() {
    // synchronize when either:
    //     - read has almost caught up to or is ahead of write, or
    //     - read is too far behind write
    if (_write_index_distance_traveled - _read_index_distance_traveled < (Sint64)_read_minimum_trailing_distance) {
        // read has almost caught up to write (read is too fast)
        if (!timing_is_faster()) {
            // only gather statistics for when CPU is running at normal speed
            if (!_soundbuffer_is_in_initial_sync_period()) {
                // and not in the initial sync period
                _resync_count__read_is_too_close_behind_write++;
            }
        }
        _soundbuffer_rewind_read();
    }
    else if (_write_index_distance_traveled - _read_index_distance_traveled > (Sint64)_read_maximum_trailing_distance) {
        // read is too far behind write (read is too slow)
        if (!timing_is_faster()) {
            // only gather statistics for when CPU is running at normal speed
            if (!_soundbuffer_is_in_initial_sync_period()) {
                // and not in the initial sync period
                _resync_count__read_is_too_far_behind_write++;
            }
        }
        _soundbuffer_fastforward_read();
    }

    // perform the actual read
    Uint8 value = _buffer[_read_index++];
    _read_index_distance_traveled++;
    if (_read_index == _bufferSize) {
        _read_index = 0;
    }

    // debug
    /*sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "read:%d  write:%d      read_distance:%lld  write_distance:%lld",
        _circular_read_index,
        _circular_write_index,
        _circular_read_index_distance_traveled,
        _circular_write_index_distance_traveled);
    log_write((char*)_debugBuffer);*/

    return value;
}

void soundbuffer_write(Uint8 value) {
    _buffer[_write_index++] = value;
    _write_index_distance_traveled++;
    if (_write_index == _bufferSize) {
        _write_index = 0;
    }
}

void soundbuffer_start(Uint32 sampleRatePerSecond) {
    _sampleRatePerSecond = sampleRatePerSecond;
    _bufferSize = _sampleRatePerSecond / 3;
    _buffer = (Uint8*)malloc(_bufferSize * sizeof(Uint8*));
    if (_buffer == NULL) {
        log_write("Error: could not allocate memory for the audio circular buffer");
        return;
    }
    memset(_buffer, 0, _bufferSize);
    _read_index = 0;
    _write_index = 0;

    _read_index_distance_traveled = 0;
    _write_index_distance_traveled = 0;

    _resync_count__read_is_too_close_behind_write = 0;
    _resync_count__read_is_too_far_behind_write = 0;

    _read_minimum_trailing_distance = _sampleRatePerSecond * READ_MINIMUM_TRAILING_DISTANCE__FACTOR / 3000;
    _read_maximum_trailing_distance = _sampleRatePerSecond * READ_MAXIMUM_TRAILING_DISTANCE__FACTOR / 3000;
    _read_sync_rewind_amount = _sampleRatePerSecond * READ_SYNC_REWIND_AMOUNT__FACTOR / 3000;
    _read_sync_fastforward_amount = _sampleRatePerSecond * READ_SYNC_FASTFORWARD_AMOUNT__FACTOR / 3000;
}

void soundbuffer_destroy() {
    if (_buffer != NULL) {
        free(_buffer);
        _buffer = NULL;
    }

    // time spent reading
    double seconds = (double)_read_index_distance_traveled / (double)_sampleRatePerSecond;
    // we don't report resyncs during the sync period
    seconds -= _initialSyncPeriodSeconds;

    double readTooClosePerSecond = 0.0f;
    double readTooFarPerSecond = 0.0f;

    if (seconds > 0.0f) {
        readTooClosePerSecond = (double)_resync_count__read_is_too_close_behind_write / seconds;
        readTooFarPerSecond = (double)_resync_count__read_is_too_far_behind_write / seconds;
    }
    
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Sound resync (read too fast): %d (%.1f/sec)",
        _resync_count__read_is_too_close_behind_write,
        readTooClosePerSecond);
    log_write((char*)_debugBuffer);

    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Sound resync (read too slow): %d (%.1f/sec)",
        _resync_count__read_is_too_far_behind_write,
        readTooFarPerSecond);
    log_write((char*)_debugBuffer);

    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Sample rate: %d/sec",
        _sampleRatePerSecond);
    log_write((char*)_debugBuffer);
    
    double bufferSizeSeconds = (double)_bufferSize / (double)_sampleRatePerSecond;
    sprintf_s((char* const)_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Buffer size: %.2fsec (%d samples)",
        bufferSizeSeconds,
        _bufferSize);
    log_write((char*)_debugBuffer);
}
